/*
 * Everything from here will be injected to the WebApp Window
 * after the DOM is loaded, but before anything is executed.
 * For now only inject something when are at the toplevel window.
 */
if (window.parent === window) {
	console.log("Injecting to toplevel window");
	/*
	 * Get the nw window object. We unfortunately have to mess around
	 * with the require function in favor of WebMeetings require namespace.
	 * The requirejs namespace could be used by WebMeetings instead of require.
	 */
	window.requireJS = window.require;
	window.require = window.requireNode;
	var win = nw.Window.get();
	window.requireNode = window.require;
	window.require = window.requireJS;

	var fs, p;

	/*
	 * If Zarafa is defined webapp is being loaded.
	 * This means we can add the custom context menus and
	 * and a listener for paralyzed connections to relogin.
	 */
	if (typeof Zarafa !== 'undefined') {
		addContextMenus();
		fs = window.requireNode('fs');
		p = window.requireNode('path');
		/* Remove the favicon from webapp, and let deskapp use its own */
		document.querySelector("link[rel*='shortcut icon']").remove();
		document.querySelector("link[rel*='icon']").remove();

		var favicon = document.createElement('link');
		favicon.rel = 'icon';
		favicon.href = chrome.runtime.getURL('app/resources/images/logo.png');
		favicon.type = 'image/png';
		favicon.sizes = ['256x256'];
		document.head.appendChild(favicon);

		Zarafa.onReady(function() {
			console.log("WebApp is ready");
			/* Add event listener for paralyzed connections */
			window.container.getRequest().on('connectionparalyzed', function(req, reason) {
				if (reason == window.Zarafa.core.data.ParalyzeReason.SESSION_EXPIRED || reason == window.Zarafa
					.core.data.ParalyzeReason.SESSION_INVALID) {
					win.customEventEmitter.emit('WAparalyzed');
				}
			});
			/* EventListener for webapps reload event. */
			container.addListener('beforewebappreload', function() {
				win.customEventEmitter.emit('WAreload');
			});
			/* EventListener for webapps loaded event. Not used yet.*/
			container.addListener('webapploaded', function() {
				console.log('webapploaded fired.');
			});
			overrideTitle();
			/* Add functions for creating mails, appointments, etc. */
			win.newItem = newItem;
			win.createMail = createMail;
			win.openEML = openEML;
			/* Add function for opening windows by webapp */
			window.deskappOpenWindow = deskappOpenWindow;
		});
	}

	/* Observe loading-mask removement. This means webapp is fully loaded */
	var loadObserver = new MutationObserver(function(mutations) {
		mutations.forEach(function(mutation) {
			if (!mutation.removedNodes) {
				return;
			}
			for (var i = 0; i < mutation.removedNodes.length; i++) {
				if (mutation.removedNodes[i].id === 'loading-mask') {
					/* Skip the welcome page */
					if (document.getElementsByClassName('zarafa-welcome-body').length > 0) {
						document.getElementsByTagName('button')[0].click();
						return;
					}
					/* Add event listener for logout button */
					var logoutButton = document.getElementById('mainmenu-button-logout');
					logoutButton.addEventListener('click', function() {
						window.container.logout();
						win.customEventEmitter.emit('WAlogout');
					});
					/* Tell main window we are ready to be shown */
					win.customEventEmitter.emit('WAloaded');
					/* Stop observing */
					loadObserver.disconnect();
				}
			}
		});
	});
	loadObserver.observe(document.body, {
		childList: true,
		subtree: true,
		attributes: false,
		characterData: false
	});
}

/**
 * Basically replace WebApp by DeskApp in the title.
 * If the titlecounter plugin is enabled we need to add an
 * mutation observer to catch the updates from it.
 * If it is not enabled we just override the title once.
 */
function overrideTitle() {
	if (container.getSettingsModel().get('zarafa/v1/plugins/titlecounter/enable')) {
		var title = document.querySelector('head > title');
		var titleObserver = new MutationObserver(function(mutations) {
			mutations.forEach(function(mutation) {
				if (mutation.target.textContent.search(/webapp/i) !== -1) {
					document.title = mutation.target.textContent.replace(/webapp/gi, 'DeskApp');
				}
			});
		});
		titleObserver.observe(title, {
			subtree: true
			, characterData: true
			, childList: true
		});
	} else {
		document.title = document.title.replace(/webapp/gi, 'DeskApp');
	}
}

/**
 * With this function webapp can open a new nw Window and get back
 * a window object.
 * @param {String} url url to open
 * @param {Function} callback callback function with the window object
 * @param {Object} options for the window to open, optional
 * @return {Function} return the callback
 */
function deskappOpenWindow(url, callback, options) {
	/* For backwards compatibility set options if there are none */
	if (!options) {
		options = {
			'title': 'Kopano DeskApp',
			'width': 950,
			'height': 600,
			'position': 'center'
		};
	}

	/* Always open the window hidden */
	options.show = false;

	/* Injection scripts have changed their location */
	if (options.inject_js_start) {
		options.inject_js_start = 'app/injection/' + options.inject_js_start;
	}
	if (options.inject_js_end) {
		options.inject_js_end = 'app/injection/' + options.inject_js_end;
	}

	nw.Window.open(url, options, function(newWindow) {
		if (newWindow) {
			/* wait until the window has loaded, before manipulating and showing it */
			newWindow.on('loaded', function() {
				if (options.x) {
					newWindow.x = options.x;
				}
				if (options.y) {
					newWindow.y = options.y;
				}
				newWindow.window.document.title = newWindow.window.document.title.replace(/webapp/gi, 'DeskApp');
				newWindow.show();
			});
			newWindow.on('new-win-policy', function(frame, url, policy) {
				if (url) {
					nw.Shell.openExternal(url);
					policy.ignore();
				}
			});
			newWindow.on('enter-fullscreen', function() {
				// global shortcut to leave fullscreen
				var escapeShortcut = new nw.Shortcut({
					key: "Escape",
					active: function () {
						if (newWindow) {
							newWindow.leaveFullscreen();
							nw.App.unregisterGlobalHotKey(escapeShortcut);
						}
					}
				});
				nw.App.registerGlobalHotKey(escapeShortcut);
			});
			return (callback(newWindow.window));
		} else {
			/* opening the Window went wrong, so return null */
			return (callback(null));
		}
	});
}

/**
 * This adds our custom context menu items.
 */
function addContextMenus() {
	/*
	 * The onClicked callback function is executed when
	 * a context menu item is clicked.
	 * @info: object, information sent when a context menu item is clicked.
	 */
	function onContextHandler(info) {
		switch (info.menuItemId) {
			case "print_selection":
				printText(info["selectionText"]);
				break;
			default:
				console.log("Unknown contextmenu item. Do nothing");
				break;
		}
	}

	// Add the listener for clicked context menu items
	chrome.contextMenus.onClicked.addListener(onContextHandler);

	// Set up the different context menu items
	chrome.contextMenus.removeAll(function() {
		chrome.contextMenus.create({
			"title": global.gt.gettext("Print\u2026"),
			"contexts": ["selection", "frame"],
			"id": "print_selection"
		});
	});

	/**
	 * Print selection, or the whole iframe from where print_selection
	 * context menu item has been called.
	 * If called from a selection everything is copied to an iframe,
	 * which is then printed. If nothing is selected print the whole
	 * iframe from where we have been called.
	 * @param {String} text text to print
	 */
	function printText(text) {
		var activeElem = document.activeElement;
		var printWin = activeElem.contentWindow;
		var sel = printWin.getSelection();
		if (text && sel.rangeCount) {
			var printFrame = document.createElement('iframe');
			printFrame.style.cssText = "height: 0px; width: 0px; position: absolute;";
			document.body.appendChild(printFrame);
			var contentW = printFrame.contentWindow;
			for (var i = 0, len = sel.rangeCount; i < len; ++i) {
				contentW.document.body.appendChild(sel.getRangeAt(i).cloneContents());
			}
			contentW.print();
			printFrame.remove();
		} else {
			printWin.print();
		}
	}
}
/**
 * Create a new item of @type {appointment, meeting, note, contact, task}.
 * @param {String} type type of the new Item to create
 */
function newItem(type) {
	var isMeeting;
	// appointment and meeting have the same context calendar, but differ in creation
	isMeeting = false;
	if (type === 'appointment') {
		type = 'calendar';
	} else if (type === 'meeting') {
		type = 'calendar';
		isMeeting = true;
	}
	var model = window.container.getContextByName(type).getModel();
	var record = model.createRecord();
	if (isMeeting) {
		record.convertToMeeting();
	}
	/* show the created item */
	Zarafa.core.data.UIFactory.openCreateRecord(record, {
		model: model
	});
	win.show();
	win.focus();
}

/**
 * This creates a new E-Mail.
 * @param {Object} options parsed parameters from getopt
 */
function createMail(options) {
	var sSubject;

	var model = container.getContextByName('mail').getModel();
	var record = model.createRecord();
	record.beginEdit();

	if (options.subject) {
		sSubject = decodeURIComponent(options.subject);
	}

	/*
	 * Convert attachment array to file list.
	 * Check if the files exist and if the file is a directory.
	 * Zip directories to the OSs temporary path, then add it to the mail.
	 */
	if (options.attach) {
		var files = new FileList();
		var attachments = options.attach;
		var stats;
		for (var i = 0; i < attachments.length; ++i) {
			var fullpath = decodeURIComponent(attachments[i]).replace('file://', '');
			/* Be sure the file exists */
			try {
				stats = fs.statSync(fullpath);
			} catch (e) {
				continue;
			}
			var path = p.parse(fullpath);
			var filename;
			/* If we get a directory zip it */
			if (stats.isDirectory()) {
				filename = path.name + '.zip';
				var target = p.join(window.requireNode('os').tmpdir(), filename);
				zipDirectory(path, target, record);
			} else {
				filename = path.base;
				files.append(new File(fullpath, filename));
			}
			if (!options.subject) {
				if (!sSubject) {
					sSubject = global.gt.gettext('Sending as attachment(s): ');
					sSubject += filename;
				} else {
					sSubject += ', ' + filename;
				}
			}
		}
		var store = record.getSubStore('attachments');
		store.uploadFiles(files);
	}

	if (options.body) {
		var sBody = decodeURIComponent(options.body);
		record.set('isHTML', false);
		record.set('body', sBody);
		record.set('html_body', Ext.util.Format.htmlEncode(sBody));
	}

	if (sSubject) {
		record.set('subject', sSubject);
	}

	/* If there are any recipients, then add them to the recipient store */
	if (options.to || options.cc || options.bcc) {
		var recipientStore = record.getRecipientStore();
		if (options.to) {
			addRecipients(recipientStore, options.to, Zarafa.core.mapi.RecipientType.MAPI_TO);
		}
		if (options.cc) {
			addRecipients(recipientStore, options.cc, Zarafa.core.mapi.RecipientType.MAPI_CC);
		}
		if (options.bcc) {
			addRecipients(recipientStore, options.bcc, Zarafa.core.mapi.RecipientType.MAPI_BCC);
		}
	}

	record.endEdit();

	/* show the created mail */
	Zarafa.core.data.UIFactory.openCreateRecord(record, {
		browser: false,
		model: model
	});
	win.show();
	win.focus();
}

/**
 * Add recipients to the records recipient store
 * @param {Object} recipientStore mail recipientStore
 * @param {Array} aRec the recipients to add
 * @param {Zarafa.core.mapi.RecipientType} type the recipient type (MAPI_TO, MAPI_CC, MAPI_BCC)
 */
function addRecipients(recipientStore, aRec, type) {
	var match, smtp, name;
	var regex = /<([^<>]*)>/g;
	for (var i = 0; i < aRec.length; ++i) {
		var fulladdress = aRec[i];
		if ((match = regex.exec(fulladdress)) != null) {
			smtp = match[1] ? match[1] : match[0];
			name = fulladdress.replace('<' + smtp + '>', '').trim();
		} else {
			smtp = fulladdress;
			name = smtp;
		}
		var rec = createRecipient(smtp, name, type);
		recipientStore.add(rec);
	}
}

/**
 * Create a recipient object
 * @param {String} smtp the smtp address
 * @param {String} name display name
 * @param {Zarafa.core.mapi.RecipientType} type the recipient type (MAPI_TO, MAPI_CC, MAPI_BCC)
 * @return {Record} record holding the recipient
 */
function createRecipient(smtp, name, type) {
	var props = {
		object_type: Zarafa.core.mapi.ObjectType.MAPI_MAILUSER,
		recipient_type: type,
		smtp_address: smtp,
		display_name: name,
		address_type: 'SMTP'
	};
	return Zarafa.core.data.RecordFactory.createRecordObjectByCustomType(
		Zarafa.core.data.RecordCustomObjectType.ZARAFA_RECIPIENT, props);
}

/**
 * Create a zipfile from a directory and attach it to a mail record
 * @param {Object} path parsed path Object of the directory to be zipped
 * @param {String} target full filename to zip file
 * @param {Object} record mail record to add the target to
 */
function zipDirectory(path, target, record) {
	var zipper = window.requireNode("adm-zip");
	var dirname = path.name;
	var fullpath = p.join(path.dir, path.base);
	var zip = new zipper();
	var notifier = container.getNotifier();
	var curSize = 0;
	var notification;

	notification = notifier.notify('info', global.gt.gettext('Preparing attachment'),
		global.gt.gettext('Archiving directory ') + fullpath + ".", {
			persistent: true
		});
	zip.addLocalFolder(fullpath);
	zip.toBuffer(
		/* On success */
		function(buf) {
			fs.writeFileSync(target, buf);
			notifier.notify('info', null, null, {
				reference: notification,
				destroy: true
			});
			notifier.notify('info', global.gt.gettext('Attachment prepared'),
				fileSize(curSize) + global.gt.gettext(' archived. Attaching\u2026'));
			record.beginEdit();
			var files = new FileList();
			files.append(new File(target, dirname + '.zip'));
			var store = record.getSubStore('attachments');
			store.uploadFiles(files);
			record.endEdit();
		},
		/* On error */
		function(err) {
			notifier.notify('info', null, null, {
				reference: notification,
				destroy: true
			});
			notifier.notify('error', global.gt.gettext('Error zipping Folder'),
				global.gt.gettext('An error has occured zipping folder ') + fullpath * '.');
		},
		/* On fileStart */
		function(fileName) {
			notifier.notify(
				'info', global.gt.gettext('Preparing attachment'),
				global.gt.gettext('Adding files. Current size: ') + fileSize(curSize) + '.', {
					reference: notification,
					update: true
				}
			);
		},
		/* On fileend */
		function(fileName) {
			var stats = fs.statSync(fullpath + p.sep + fileName);
			if (!stats.isDirectory()) {
				curSize += stats.size;
			}
		}
	);
}

/**
 * Convert bytes to a human readable format.
 * @param {String} bytes bytes to convert
 * @return {String} bytes in human readable format
 */
function fileSize(bytes) {
	var exp = Math.log(bytes) / Math.log(1024) | 0;
	var result = (bytes / Math.pow(1024, exp)).toFixed(2);

	return result + ' ' + (exp == 0 ? 'bytes' : 'KMGTPEZY' [exp - 1] + 'B');
}

/**
 * Open an eml in WebApp
 * @param {String} eml full path to eml file
 */

function openEML(eml) {
	if (eml === '') {
		return;
	}
	if (!(Ext.isFunction(Zarafa.common.Actions.importEmlCallback) || Ext.isFunction(Zarafa.common.Actions.importItemCallback))){
		return;
	}
	var fullpath = decodeURIComponent(eml).replace('file://', '');
	/* Be sure the file exists */
	try {
		stats = fs.statSync(fullpath);
	} catch (e) {
		var notifier = container.getNotifier();
		notifier.notify('error', global.gt.gettext('Error opening EML'),
			global.gt.gettext('The file') + ' ' + fullpath + ' ' + global.gt.gettext('does not exist.'));
		return;
        }
	var folder = container.getHierarchyStore().getDefaultFolder('drafts');
	var files = new FileList();
	var path = p.parse(fullpath);
	files.append(new File(fullpath, path.base));
	if (Ext.isFunction(Zarafa.common.Actions.importItemCallback)) {
		Zarafa.common.Actions.importItemCallback(files, folder, true);
	} else {
		Zarafa.common.Actions.importEmlCallback(files, folder, true);
	}
}

/**
 * Detect Spreed WebRTC, listen to messages from spreed
 * for screensharing and take the corresponding actions.
 */
(function(body) {
	/* Detect if we are in the spreed iframe, otherwise return */
	if (!document.getElementById("globalcontext")) {
		return;
	}
	if (!body.hasAttribute("spreed-webrtc") &&
		body.getAttribute("ng-controller") !== "MediastreamController") {
		return;
	}
	console.log("Spreed WebRTC detected. Listening for messages...");
	/* Init nw's Screen API */
	nw.Screen.Init();

	/* Listen for dom message events from spreed. */
	window.addEventListener("message", function(event) {
		if (event.source !== window) {
			return;
		}

		var msg = event.data;
		if (msg.answer || msg.Type !== "Call") {
			return;
		}

		var call = msg.Call;
		if (call.Type !== "Action") {
			return;
		}

		/* Mark to avoid loops. */
		msg.answer = true;

		switch (call.Action) {
			case "chooseDesktopMedia":
				nw.Screen.chooseDesktopMedia(["screen", "window"], function(id) {
					if (chrome.runtime.lastError) {
						msg.Call = {
							Type: "Error",
							Error: chrome.runtime.lastError.message || chrome.runtime.lastError
						};
					} else {
						msg.Call = {
							Type: "Result",
							Result: id
						};
					}
					window.postMessage(msg, window.document.URL);
				});
				break;
			case "cancelChooseDesktopMedia":
				chrome.desktopCapture.cancelChooseDesktopMedia(call.Args);
				break;
			default:
				console.log("Unknown call action: ", call.Action);
				break;
		}
	});

	/*
	 * Webmeetings detects if the screensharing extension
	 * is available by checking for the presence of this
	 * marker element. If it is not available webmeetings
	 * shows a dialog that the extension needs to be installed.
	 */
	var marker = document.createElement("div");
	marker.id = "chromeextension-available";
	marker.className = "hidden";
	document.body.appendChild(marker);

}(document.body));
